from typing import Dict
import os
import requests as rq
import tqdm
from minio import Minio
import time

BASEURL = os.getenv("LERULI_BASEURL", "https://api.leruli.com")
SORRY = "ERROR: Accessing API failed. This is our fault, not yours. Please accept our apologies. We have been notified of this error."


def _extract_error_message(res):
    try:
        msg = res.json()["detail"]
        return msg
    except:
        pass
    try:
        msg = res.json()["detail"][0]
        param = msg["loc"][-1]
        msg = f"Parameter {param}: {msg['msg']}"
        return msg
    except:
        pass
    raise NotImplementedError()


class LeruliInternalError(Exception):
    pass


def _base_call(
    endpoint: str,
    payload: Dict,
    version: str,
    urgent: bool = False,
    progress: bool = False,
    files: Dict = {},
):
    """Common code to deal with the delayed responses as they can come from the API."""
    try:
        url = f"{BASEURL}/{version}/{endpoint}"
        if urgent is True:
            url = f"{url}?urgent={urgent}"
        res = rq.post(url, json=payload, files=files)

        pbar = None
        # wait for delayed responses
        if res.status_code == 202:
            while True:
                res = res.json()
                token = res["token"]
                remainder = res["time_to_result"]

                if progress:
                    if pbar is None:
                        pbar = tqdm.tqdm(
                            total=remainder,
                            desc="Waiting",
                            bar_format="{desc}: |{bar}| [{elapsed}<{remaining}]",
                        )
                    else:
                        pbar.total = remainder + pbar.n
                        pbar.update(0)

                early = remainder - 3
                while remainder > 0 and remainder > early:
                    tosleep = min(remainder, 1)
                    time.sleep(tosleep)
                    remainder -= 1
                    if progress:
                        pbar.update(tosleep)

                res = rq.get(
                    f"{BASEURL}/{version}/result/{token}",
                )
                if res.status_code != 202:
                    if progress:
                        pbar.total = pbar.n
                        pbar.update(0)
                        pbar.close()
                    break
            if str(res.status_code).startswith("5"):
                print(SORRY)
                raise LeruliInternalError()
            res = res.json()
        else:
            # Internal error?
            if str(res.status_code).startswith("5"):
                print(SORRY)
                raise LeruliInternalError()

            # No content?
            if res.status_code == 204:
                return None
            res = res.json()
        return res
    except:
        print(SORRY)
        raise LeruliInternalError()


def get_api_secret():
    """Fetches the current API secret from environment variables."""
    if not "LERULI_API_SECRET" in os.environ:
        print(
            "This feature is paid. Please contact info@leruli.com to obtain an API secret."
        )
        return None
    return os.getenv("LERULI_API_SECRET")


def get_group_token():
    """Fetches the current group token."""
    api_secret = get_api_secret()
    if api_secret is None:
        return None

    return _base_call("group-token", payload={"secret": api_secret}, version="latest")


def get_s3_client():
    """Builds a S3 client object from environment variables."""
    if not (
        "LERULI_S3_ACCESS" in os.environ
        and "LERULI_S3_SECRET" in os.environ
        and "LERULI_S3_SERVER" in os.environ
    ):
        print(
            "Please configure the environment variables LERULI_S3_[ACCESS|SECRET|SERVER]."
        )
        return None
    s3_access = os.getenv("LERULI_S3_ACCESS")
    s3_secret = os.getenv("LERULI_S3_SECRET")
    s3_server = os.getenv("LERULI_S3_SERVER")

    s3_client = Minio(
        s3_server.split("/")[-1],
        access_key=s3_access,
        secret_key=s3_secret,
        secure=s3_server.startswith("https"),
    )
    return s3_client



"""Interface for leruli.com"""
__author__ = "leruli.com"
__email__ = "info@leruli.com"
__version__ = "22.1.6"

from .misc import *
from .graph import *
from .task import *

__all__ = [
    "canonical_formula",
    "canonical_graph",
    "name_to_graph",
    "formula_to_graphs",
    "formula_to_cost",
    "graph_to_image",
    "graph_to_solvation_energy",
    "graph_to_geometry",
    "graph_to_name",
    "graph_to_boiling_point",
    "graph_to_melting_point",
    "graph_to_logP",
    "graph_to_logD",
    "graph_to_pKa",
    "graph_to_formula",
    "get_s3_client",
    "get_api_secret",
    "task_submit",
    "task_status",
    "task_get",
    "task_cancel",
    "task_publish_code",
    "task_prune",
]



"""Console script for leruli."""
import sys
import click
import leruli
import tabulate
import glob
import base64
import time as modtime

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group()
def cli1():
    pass


@cli1.command()
@click.option(
    "--format",
    type=click.Choice("XYZ PDB SDF".split(), case_sensitive=False),
    default="XYZ",
    help="File format for output, defaults to XYZ.",
)
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_geometry(graph: str, format: str, reference: str, version: str = None):
    """Molecular geometry from a graph."""
    try:
        result = leruli.graph_to_geometry(graph, format, version)
    except Exception as e:
        print(f"ERROR: {str(e)}", file=sys.stderr)
        sys.exit(1)
    if result is None:
        print("Could not find a geometry.")
    else:
        print(result["geometry"])


@click.group()
def cli2():
    pass


@cli2.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("formula")
def canonical_formula(
    formula: str,
    reference: bool,
    version: str,
):
    """Canonicalize a sum formula."""
    result = leruli.canonical_formula(formula, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["formula"])


@click.group()
def cli5():
    pass


@cli5.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_formula(
    graph: str,
    reference: bool,
    version: str,
):
    """Obtain a formula from a graph."""
    result = leruli.graph_to_formula(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["formula"])


@click.group()
def cli20():
    pass


@cli20.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_boiling_point(
    graph: str,
    reference: bool,
    version: str,
):
    """Estimate a boiling point in deg C."""
    result = leruli.graph_to_boiling_point(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["bp"])


@click.group()
def cli21():
    pass


@cli21.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_melting_point(
    graph: str,
    reference: bool,
    version: str,
):
    """Estimate a melting point in deg C."""
    result = leruli.graph_to_melting_point(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["mp"])


@click.group()
def cli22():
    pass


@cli22.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_logP(
    graph: str,
    reference: bool,
    version: str,
):
    """Estimate a logP."""
    result = leruli.graph_to_logP(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["logp"])


@click.group()
def cli23():
    pass


@cli23.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_logD(
    graph: str,
    reference: bool,
    version: str,
):
    """Estimate a logD."""
    result = leruli.graph_to_logD(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["logp"])


@click.group()
def cli24():
    pass


@cli24.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_pKa(
    graph: str,
    reference: bool,
    version: str,
):
    """Estimate a pKa."""
    result = leruli.graph_to_pKa(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result)


@click.group()
def cli3():
    pass


@cli3.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def canonical_graph(
    graph: str,
    reference: bool,
    version: str,
):
    """Canonicalize a graph."""
    result = leruli.canonical_graph(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["graph"])


@click.group()
def cli6():
    pass


@cli6.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("formula")
def formula_to_graphs(
    formula: str,
    reference: bool,
    version: str,
):
    """Find molecular graphs of a given sum formula."""
    result = leruli.formula_to_graphs(formula, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    for molecule in result["molecules"]:
        print(f"{molecule['name']}: {molecule['smiles']}")


@click.group()
def cli7():
    pass


@cli7.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
def graph_to_name(
    graph: str,
    reference: bool,
    version: str,
):
    """Obtain a name from a graph."""
    result = leruli.graph_to_name(graph, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["name"])


@click.group()
def cli8():
    pass


@cli8.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("name")
def name_to_graph(
    name: str,
    reference: bool,
    version: str,
):
    """Obtain a graph from a name."""
    result = leruli.name_to_graph(name, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(result["graph"])


@click.group()
def cli9():
    pass


@cli9.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("formula")
@click.argument("basisset")
def formula_to_cost(
    formula: str,
    basisset: str,
    reference: bool,
    version: str,
):
    """Obtain a cost estimate for a molecule."""
    result = leruli.formula_to_cost(formula, basisset, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    print(tabulate.tabulate(result["data"], headers="keys"))


@click.group()
def cli10():
    pass


@cli10.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option("--angle", default=0, help="Rotation angle.")
@click.option(
    "--format",
    type=click.Choice("SVG PNG PDF".split(), case_sensitive=False),
    default="SVG",
    help="File format for output, defaults to SVG.",
)
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.argument("graph")
@click.argument("filename")
def graph_to_image(
    filename: str,
    graph: str,
    format: str,
    angle: int,
    reference: bool,
    version: str,
):
    """Structure formula as image file."""
    result = leruli.graph_to_image(graph, format, angle, version)
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )
    with open(filename, "wb") as fh:
        fh.write(base64.b64decode(result["image"]))


@click.group()
def cli11():
    pass


@cli11.command()
@click.option("--version", default="latest", help="Request specific API version.")
@click.option(
    "--reference", is_flag=True, default=False, help="Print references for the result."
)
@click.option("--quiet", is_flag=True, default=False, help="Hide progress bar.")
@click.argument("graph")
@click.argument("solvent")
@click.argument("temperatures")
def graph_to_solvation_energy(
    graph: str,
    solvent: str,
    temperatures: str,
    reference: bool,
    version: str,
    quiet: bool,
):
    """Estimate the free energy of solvation."""
    temperatures = [float(_) for _ in temperatures.split(",")]
    result = leruli.graph_to_solvation_energy(
        graph, solvent, temperatures, version, progress=not quiet
    )
    if reference:
        print(
            f"# Reference as BibTeX: https://api.leruli.com/{version}/references/{result['reference']}/bibtex"
        )

    if "detail" in result:
        print(f"ERROR: {result['detail']['ERROR']}")
    else:
        printable = [
            {"Temperature [K]": k, "Energy of solvation [kcal/mol]": v}
            for k, v in result["solvation_energies"].items()
        ]
        print(tabulate.tabulate(printable, headers="keys"))


@click.group()
def group_task_submit():
    pass


@group_task_submit.command()
@click.option("--memory", default=4000, help="Memory limit in MB.")
@click.option("--time", default=60 * 24, help="Time limit in minutes.")
@click.option("--cores", default=1, help="Number of cores to allocate.")
@click.option(
    "--batch",
    nargs=1,
    help="Submit multiple directories at once, specified as either a string of a globbing pattern (e.g., case-?/run-*/) or a file with a line-by-line list of directories.",
)
@click.argument("code")
@click.argument("version")
@click.argument("command", nargs=-1, required=True)
def task_submit(
    memory: int,
    time: int,
    cores: int,
    batch: str,
    code: str,
    version: str,
    command: str,
):
    """Submit one or many jobs to the Leruli queue. This is a paid feature requiring an API secret, which can be obtained from info@leruli.com."""

    if batch is None:
        try:
            jobid = leruli.task_submit(
                ".", code, version, command, cores, memory, time * 60
            )
        except ValueError as e:
            print(f"Not submitted: {str(e)}")
            sys.exit(1)
        if jobid is not None:
            print(jobid)
        else:
            sys.exit(1)
    else:
        try:
            with open(batch) as fh:
                directories = []
                for line in fh.readlines():
                    line = line.strip()
                    if len(line) > 0:
                        directories.append(line)
        except:
            directories = glob.glob(batch)

        njobs = len(directories)
        starttime = modtime.time()
        failed = leruli.task_submit_many(
            directories,
            [code] * njobs,
            [version] * njobs,
            [command] * njobs,
            [cores] * njobs,
            [memory] * njobs,
            [time * 60] * njobs,
        )
        stoptime = modtime.time()
        for directory in sorted(failed.keys()):
            print(f"Not submitted: {directory} - {failed[directory]}")
        duration = stoptime - starttime
        print(f"Submitted {njobs} in {duration:1.1f}s ({njobs/duration:1.0f}/s)")
        if len(failed) > 0:
            sys.exit(1)


@click.group()
def group_task_status():
    pass


@group_task_status.command()
@click.argument("jobid", required=False)
def task_status(jobid: str):
    """Get the status of a job in the queue."""
    if jobid is None:
        try:
            with open("leruli.job") as fh:
                jobid = fh.read().strip()
        except FileNotFoundError:
            print ("No job ID specified and no leruli.job in current folder.")
            sys.exit(1)
    try:
        status = leruli.task_status(jobid)
    except ValueError:
        print ("No such job or not your job.")
        sys.exit(1)
    if status is None:
        sys.exit(1)
    if status["reason"] is None:
        print(status["status"])
    else:
        print(status["status"] + ": " + status["reason"])


@click.group()
def group_task_get():
    pass


@group_task_get.command()
@click.argument("bucket", required=False)
def task_get(bucket: str):
    """Retrieve the input and result files of a job."""
    if bucket is None:
        with open("leruli.bucket") as fh:
            bucket = fh.read().strip()
    leruli.task_get(".", bucket)


@click.group()
def group_task_cancel():
    pass


@group_task_cancel.command()
@click.argument("jobid", required=False)
def task_cancel(jobid: str):
    """Cancels a job in the queue."""
    if jobid is None:
        with open("leruli.job") as fh:
            jobid = fh.read().strip()
    status = leruli.task_cancel(jobid)
    if status is None:
        sys.exit(1)
    if status["reason"] is None:
        print(status["status"])
    else:
        print(status["status"] + ": " + status["reason"])


@click.group()
def group_task_list_codes():
    pass


@group_task_list_codes.command()
def task_list_codes():
    """Lists all codes available to you."""
    codes = leruli.task_list_codes()
    for code, version in codes:
        print(f"{code}:{version}")


@click.group()
def group_task_prune():
    pass


@group_task_prune.command()
@click.argument("jobid", required=False)
@click.argument("bucket", required=False)
def task_prune(jobid: str, bucket: str):
    """Deletes the input/output files of a bucket."""
    if (jobid is None and bucket is not None) or (jobid is not None and bucket is None):
        print("Either specify none or both of jobid and bucket.")
        sys.exit(1)

    if jobid is None:
        with open("leruli.job") as fh:
            jobid = fh.read().strip()
    status = leruli.task_status(jobid)
    if status["status"] == "running":
        print("Cannot prune a running job: the job would complete and fail.")
        sys.exit(1)

    if bucket is None:
        with open("leruli.bucket") as fh:
            bucket = fh.read().strip()
    try:
        leruli.task_prune(bucket)
    except:
        print("Pruning failed. Bucket already pruned or wrong S3 credentials.")
        sys.exit(1)


@click.group()
def group_task_publish_code():
    pass


@group_task_publish_code.command()
@click.argument("code")
@click.argument("version")
def task_publish_code(code: str, version: str):
    """Uploads a local docker image to the queue."""
    leruli.task_publish_code(code, version)


cli = click.CommandCollection(
    sources=[
        cli1,
        cli2,
        cli3,
        cli20,
        cli21,
        cli22,
        cli23,
        cli24,
        cli5,
        cli6,
        cli7,
        cli8,
        cli9,
        cli10,
        cli11,
        group_task_submit,
        group_task_status,
        # group_task_cancel,
        group_task_publish_code,
        group_task_prune,
        group_task_get,
        group_task_list_codes,
        group_task_cancel,
    ],
    context_settings=CONTEXT_SETTINGS,
)


def main():
    sys.exit(cli())


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = ["Click>=7.0", "requests", "tabulate", "minio", "docker", "tqdm>=4.62.0", "aiobotocore", "aiohttp"]

test_requirements = [
    "pytest>=3",
]

setup(
    author="leruli.com",
    author_email="info@leruli.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Use leruli.com from command line and Python code.",
    entry_points={
        "console_scripts": [
            "leruli=leruli.cli:main",
        ],
    },
    install_requires=requirements,
    long_description=readme,
    include_package_data=True,
    keywords="leruli",
    name="leruli",
    packages=find_packages(include=["leruli", "leruli.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    version="22.1.6",
    zip_safe=False,
)



from . import internal
from typing import List


def graph_to_image(
    graph: str,
    format: str,
    angle: int,
    version: str = "latest",
    urgent: bool = False,
    progress: bool = False,
):
    payload = {"graph": graph, "format": format, "angle": angle}
    return internal._base_call("graph-to-image", payload, version, urgent, progress)


def graph_to_solvation_energy(
    graph: str,
    solventname: str,
    temperatures: List[float],
    version: str = "latest",
    urgent: bool = False,
    progress: bool = False,
):
    payload = {"graph": graph, "solvent": solventname, "temperatures": temperatures}
    return internal._base_call(
        "graph-to-solvation-energy", payload, version, urgent, progress
    )


def graph_to_geometry(
    graph: str,
    format: str,
    version: str = "latest",
    urgent: bool = False,
    progress: bool = False,
):
    """Obtains a 3D geometry for a molecular graph.

    Parameters
    ----------
    graph : str
        SMILES
    format : str
        The requested geometry format: XYZ, SDF, PDB.
    version : str, optional
        Specific version of the API to use, by default "latest"
    urgent : bool, optional
        For interactive use only, by default False
    progress : bool, optional
        Whether to show a progress bar, by default False

    Returns
    -------
    str or None
        String of the file, None if no geometry could be found
    """
    payload = {"graph": graph, "format": format}
    return internal._base_call("graph-to-geometry", payload, version, urgent, progress)


def graph_to_name(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("graph-to-name", payload, version, urgent, progress)


def graph_to_boiling_point(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call(
        "graph-to-boiling-point", payload, version, urgent, progress
    )


def graph_to_melting_point(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call(
        "graph-to-melting-point", payload, version, urgent, progress
    )


def graph_to_logP(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("graph-to-logP", payload, version, urgent, progress)


def graph_to_logD(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("graph-to-logD", payload, version, urgent, progress)


def graph_to_pKa(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("graph-to-pKa", payload, version, urgent, progress)


def graph_to_formula(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("graph-to-formula", payload, version, urgent, progress)



import tarfile
import os.path

def make_tarfile(output_filename, source_dir):
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.basename(source_dir))
make_tarfile("../test.tgz", ".")



import aiohttp
from . import internal
import os
import io
import tarfile
import gzip
import docker
import requests as rq
import uuid
from typing import List, Iterable
import tqdm
import tqdm.asyncio
import asyncio
import aiobotocore.session
import multiprocessing as mp
import time


async def _async_object_stage(
    s3_client,
    directory: str,
    code: str,
    version: str,
    command: str,
    cores: int,
    memorymb: int,
    timeseconds: int,
):
    global counter
    global rank

    api_secret = internal.get_api_secret()
    if api_secret is None:
        counter[rank] += 2
        return {"error": "No API token configured", "directory": directory}

    if os.path.exists(f"{directory}/leruli.job"):
        counter[rank] += 2
        return {"error": "Directory already submitted.", "directory": directory}

    # TODO: detect failed S3 interactions
    bucket = str(uuid.uuid4())
    await s3_client.create_bucket(Bucket=bucket)
    counter[rank] += 1

    # in-memory tar file
    buffer = io.BytesIO()
    with tarfile.open(fileobj=buffer, mode="w:gz") as tar:
        tar.add(directory, arcname=os.path.basename("."))

        runscript = io.BytesIO(("#!/bin/bash\n" + " ".join(command)).encode("ascii"))
        tarinfo = tarfile.TarInfo(name="run.sh")
        tarinfo.size = runscript.getbuffer().nbytes
        tar.addfile(tarinfo=tarinfo, fileobj=runscript)
    buffer.seek(0)

    # upload
    await s3_client.put_object(Bucket=bucket, Key="run.tgz", Body=buffer)

    # submit to API
    codeversion = f"{code}:{version}"
    payload = {
        "secret": api_secret,
        "bucketid": bucket,
        "name": "default",
        "codeversion": codeversion,
        "cores": cores,
        "memorymb": memorymb,
        "timelimit": timeseconds,
        "directory": directory,
    }

    counter[rank] += 1
    return payload


def task_submit_many(
    directories: Iterable[str],
    codes: Iterable[str],
    versions: Iterable[str],
    commands: Iterable[str],
    cores: Iterable[int],
    memorymb: Iterable[int],
    timeseconds: Iterable[int],
) -> List[str]:
    """Submits many calculations to Leruli Queue/Cloud at once.

    Parameters
    ----------
    directories : Iterable[str]
        Paths to directories, either absolute or relative.
    codes : Iterable[str]
        Codes to run. Needs to be of same length as `directories`.
    versions : Iterable[str]
        Code versions to use. Needs to be of same length as `directories`.
    commands : Iterable[str]
        Commands to execute. Needs to be of same length as `directories`.
    cores : Iterable[int]
        Cores to use. Needs to be of same length as `directories`.
    memorymb : Iterable[int]
        Memory to use in MB. Needs to be of same length as `directories`.
    timeseconds : Iterable[int]
        Time limit for jobs in seconds. Needs to be of same length as `directories`.

    Returns
    -------
    dict[str]
        Keys: directories which could not be submitted, values: reasons for this to happen.
    """

    cases = list(
        zip(directories, codes, versions, commands, cores, memorymb, timeseconds)
    )

    def split(a, n):
        k, m = divmod(len(a), n)
        return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))

    # TODO find number of procs
    nprocs = 5
    segments = split(cases, nprocs)

    # assign one id to every worker
    idqueue = mp.Queue()
    [idqueue.put(_) for _ in range(nprocs)]
    progress_counter = mp.Array("i", nprocs, lock=False)

    # Uploading
    with tqdm.tqdm(total=len(cases), desc="Uploading job data") as pbar:
        with mp.Pool(
            nprocs,
            initializer=_task_submit_many_initializer,
            initargs=(idqueue, progress_counter),
        ) as p:
            result = p.map_async(_task_submit_many_toasync, segments)
            while True:
                total = sum([progress_counter[_] for _ in range(nprocs)])
                pbar.n = int(total / 2)
                pbar.refresh()
                try:
                    result.successful()
                except:
                    time.sleep(0.1)
                    continue
                result.wait()
                break

        # prepare for stage 2: submission to bulk API
        failed = {}
        cases = []
        for case in sum(result.get(), []):
            if "error" in case:
                failed[case["directory"]] = case["error"]
            else:
                payload = case.copy()
                del payload["directory"]
                cases.append((payload, case["directory"]))

    results = asyncio.run(_task_submit_many_API_toasync(cases))
    for result in results:
        failed.update(results)
    return failed


async def _task_submit_many_API_toasync(cases):
    # build parallel task list
    async with aiohttp.ClientSession() as session:
        tasks = []
        while len(cases) > 0:
            segment, cases = cases[:50], cases[50:]
            tasks.append(_task_submit_many_API(session, segment))

        # defer for async execution
        result = await tqdm.asyncio.tqdm.gather(*tasks, desc="Submitting job batches")

    return result


def _task_submit_many_toasync(cases):
    return asyncio.run(_task_submit_many_worker(cases))


async def _task_submit_many_API(session, segment):
    failed = {}
    async with session.post(
        f"{internal.BASEURL}/v22_1/bulk/task-submit", json=[_[0] for _ in segment]
    ) as res:
        results = await res.json()
        for case, result in zip(segment, results):
            payload, directory = case

            if result["status"] != 200:
                reason = "Job not accepted"
                if result["status"] == 403:
                    reason = "API key not accepted"
                if result["status"] == 412:
                    reason = "Compute secret not set-up"
                if result["status"] == 422:
                    reason = f"Job not accepted: {str(result)}"
                failed[directory] = reason
                continue

            jobid = result["data"]
            _task_submit_finalize(directory, jobid, payload["bucketid"])
    return failed


async def _task_submit_many_worker(cases):
    session = aiobotocore.session.get_session()
    tasks = []
    async with session.create_client(
        "s3",
        endpoint_url=os.getenv("LERULI_S3_SERVER"),
        aws_secret_access_key=os.getenv("LERULI_S3_SECRET"),
        aws_access_key_id=os.getenv("LERULI_S3_ACCESS"),
    ) as s3_client:
        for args in cases:
            tasks.append(_async_object_stage(s3_client, *args))
        cases = await asyncio.gather(*tasks)

    return cases


def _task_submit_many_initializer(idqueue, progress):
    global rank
    global counter
    rank = idqueue.get()
    counter = progress

    # TODO print missing API key only once, abort immediately


def task_submit(
    directory: str,
    code: str,
    version: str,
    command: str,
    cores: int = 1,
    memorymb: int = 4000,
    timeseconds: int = 24 * 60 * 60,
):
    """Submits a given directory content as job to Leruli Queue/Cloud."""
    try:
        payload = _task_submit_payload(
            directory, code, version, command, cores, memorymb, timeseconds
        )
    except ValueError as e:
        print(f"Failed: {str(e)}")
        return

    res = rq.post(f"{internal.BASEURL}/v22_1/task-submit", json=payload)
    if res.status_code != 200:
        print("Cannot submit jobs. Please check the input.")
        return
    jobid = res.json()

    _task_submit_finalize(directory, jobid, payload["bucketid"])


def _task_submit_payload(
    directory: str,
    code: str,
    version: str,
    command: str,
    cores: int,
    memorymb: int,
    timeseconds: int,
):
    api_secret = internal.get_api_secret()
    if api_secret is None:
        return
    s3_client = internal.get_s3_client()
    if s3_client is None:
        return

    if os.path.exists(f"{directory}/leruli.job"):
        raise ValueError("Directory already submitted.")

    bucket = str(uuid.uuid4())
    s3_client.make_bucket(bucket)

    # in-memory tar file
    buffer = io.BytesIO()
    with tarfile.open(fileobj=buffer, mode="w:gz") as tar:
        tar.add(directory, arcname=os.path.basename("."))

        runscript = io.BytesIO(("#!/bin/bash\n" + " ".join(command)).encode("ascii"))
        tarinfo = tarfile.TarInfo(name="run.sh")
        tarinfo.size = runscript.getbuffer().nbytes
        tar.addfile(tarinfo=tarinfo, fileobj=runscript)
    buffer.seek(0)

    # upload
    s3_client.put_object(bucket, "run.tgz", buffer, buffer.getbuffer().nbytes)

    # submit to API
    codeversion = f"{code}:{version}"
    payload = {
        "secret": api_secret,
        "bucketid": bucket,
        "name": "default",
        "codeversion": codeversion,
        "cores": cores,
        "memorymb": memorymb,
        "timelimit": timeseconds,
    }
    return payload


def _task_submit_finalize(directory, jobid, bucket):
    # local handle
    with open(f"{directory}/leruli.job", "w") as fh:
        fh.write(f"{jobid}\n")
    with open(f"{directory}/leruli.bucket", "w") as fh:
        fh.write(f"{bucket}\n")
    return jobid


def task_status(jobid: str):
    """Queries the status of a job at Leruli Queue/Cloud."""
    api_secret = internal.get_api_secret()
    if api_secret is None:
        return

    payload = {
        "secret": api_secret,
        "jobid": jobid,
    }
    res = rq.post(f"{internal.BASEURL}/v22_1/task-status", json=payload)
    if res.status_code == 200:
        return res.json()
    if res.status_code == 404:
        raise ValueError("No such job")


def task_get(directory: str, bucket: str):
    """Downloads the input and output files of a Leruli Queue/Cloud task into a directory."""
    s3_client = internal.get_s3_client()
    for obj in s3_client.list_objects(bucket, recursive=True):
        object = obj.object_name
        dirname = os.path.dirname(object)
        if dirname != "":
            os.makedirs(dirname, exist_ok=True)
        try:
            response = s3_client.get_object(bucket, object)
            content = response.read()
        finally:
            response.close()
            response.release_conn()
        with open(f"{directory}/{object}", "wb") as fh:
            fh.write(content)


def task_cancel(jobid: str):
    """Cancels a task on Leruli Queue/Cloud."""
    api_secret = internal.get_api_secret()
    if api_secret is None:
        return

    payload = {
        "secret": api_secret,
        "jobid": jobid,
    }
    status = rq.post(f"{internal.BASEURL}/v22_1/task-cancel", json=payload)
    return status.json()


def task_publish_code(code: str, version: str):
    """Uploads a local docker image to use with Leruli Queue/Cloud."""
    api_secret = internal.get_api_secret()
    if api_secret is None:
        return
    s3_client = internal.get_s3_client()
    group = internal.get_group_token()

    client = docker.from_env()
    image = client.images.get(f"{code}:{version}")
    cache = io.BytesIO()
    for chunk in image.save(named=True):
        cache.write(chunk)
    cache.seek(0)
    tgz = gzip.compress(cache.read())
    del cache
    buffer = io.BytesIO(tgz)

    s3_client.put_object(
        f"code-{group}", f"{code}-{version}.tgz", buffer, buffer.getbuffer().nbytes
    )


def task_list_codes():
    """Lists codes available to your account on Leruli Queue/Cloud.

    Returns
    -------
    list[tuple[str, str]]
        Code name and version tuples.
    """
    api_secret = internal.get_api_secret()
    if api_secret is None:
        return []

    s3_client = internal.get_s3_client()
    group = internal.get_group_token()
    objects = s3_client.list_objects(f"code-{group}")
    codes = []
    for obj in objects:
        filename = obj.object_name
        if not filename.endswith(".tgz"):
            continue
        basename = filename[:-4]
        code, version = basename.rsplit("-", 1)
        codes.append((code, version))
    return codes


def task_prune(bucket: str):
    """Irreversibly deletes the Leruli Queue/Cloud store of input and output files."""
    s3_client = internal.get_s3_client()
    for obj in s3_client.list_objects(bucket, recursive=True):
        s3_client.remove_object(bucket, obj.object_name)
    s3_client.remove_bucket(bucket)



from . import internal


def canonical_formula(
    formula: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    """Canonicalizes a chemical (sum) formula.

    The given sum formula may contain bracketed elements, e.g. ``CH3(CH2)4CH3``.

    Parameters
    ----------
    formula : str
        Any chemical sum formula. May contain repeated elements in parentheses.
    version : str, optional
        A specific API version in case you want to enforce backwards compatibility.
    urgent : bool, optional
        Enable to call with priority, might not allow parallel execution.
    progress : bool, optional
        Enable to show interactive progress bar for this blocking call.

    Returns
    -------
    str
        Canonical sum formula in Hill notation.
    """
    payload = {"formula": formula}
    return internal._base_call("canonical-formula", payload, version, urgent, progress)


def canonical_graph(
    graph: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"graph": graph}
    return internal._base_call("canonical-graph", payload, version, urgent, progress)


def name_to_graph(
    name: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"name": name}
    return internal._base_call("name-to-graph", payload, version, urgent, progress)


def formula_to_graphs(
    formula: str, version: str = "latest", urgent: bool = False, progress: bool = False
):
    payload = {"formula": formula}
    return internal._base_call("formula-to-graphs", payload, version, urgent, progress)


def formula_to_cost(
    formula: str,
    basisset: str,
    version: str = "latest",
    urgent: bool = False,
    progress: bool = False,
):
    payload = {"formula": formula, "basisset": basisset}
    return internal._base_call("formula-to-cost", payload, version, urgent, progress)
